#include <windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#include <strsafe.h>

#ifndef M_PI
#define M_PI	3.14159265
#endif

struct D3DVERTEX {
    D3DXVECTOR3 p;
    D3DXVECTOR3 n;
    float tu, tv;
};

#define D3DFVF_D3DVERTEX (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)

IDirect3D9 *d3d;
IDirect3DDevice9 *d3ddev;

LPD3DXMESH mesh;
D3DMATERIAL9 *mesh_materials;
IDirect3DTexture9 **mesh_textures;
DWORD material_count;

IDirect3DTexture9 *rtex;
IDirect3DSurface9 *rtex_dsurf;

IDirect3DTexture9 *tex, *tex_ramp, *tex_env, *tex_back;
IDirect3DCubeTexture9 *tex_cube;
IDirect3DVertexBuffer9 *mesh_vbuf;
IDirect3DIndexBuffer9 *mesh_ibuf;
DWORD num_verts;
DWORD num_faces;

D3DXMATRIX proj_mat;  
D3DXMATRIX view_mat;
D3DXMATRIX world_mat;

bool rotate = true;
float phi = M_PI / 2.0, theta = 0, cam_dist = 20;
int sdr_idx, sdr_count;

const int width = 600, height = 600;

float global_time;

enum {
	SDR_DIFFUSE,
	SDR_TWO_LIGHTS,
	SDR_VMANIP,
	SDR_TWO_TONE,
	SDR_SPECULAR,
	SDR_TOON,
	SDR_REFL_SPH,
	SDR_REFL_CUBE,
	SDR_REFR,

	SDR_COUNT
};


IDirect3DVertexShader9 *vsdr[SDR_COUNT];
IDirect3DVertexShader9 *vsdr_cube;
IDirect3DPixelShader9 *psdr_refl, *psdr_refr;
IDirect3DPixelShader9 *post_sdr, *distort, *blur;

const char *sdr_files[] = {
	"diffuse.vs",
	"two_lights.vs",
	"vmanip.vs",
	"two_tone.vs",
	"specular.vs",
	"toon.vs",
	"refl_sph.vs",
	"refl_cube.vs",
	"refr.vs",
	0
};

// -- function prototypes --
bool init_d3d(HWND win);
bool init_geom();
void render(HWND win);
void draw_skull();
void draw_backdrop(IDirect3DTexture9 *tex, IDirect3DPixelShader9 *ps = 0);
void draw_cube();
IDirect3DVertexShader9 *load_shader(const char *fname);
IDirect3DPixelShader9 *load_pixel_shader(const char *fname);
void setup_xform();
bool create_render_target(int xsz, int ysz, IDirect3DTexture9 **rt, IDirect3DSurface9 **dsurf);
void set_render_target(IDirect3DTexture9 *rt, IDirect3DSurface9 *dsurf);
LRESULT WINAPI msg_handler(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam);


int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hprev, LPSTR cmdline, int foo)
{
    // Register the window class
    WNDCLASSEX wc = {
		sizeof(WNDCLASSEX), CS_CLASSDC, msg_handler, 0L, 0L, 
		hinst, NULL, NULL, NULL, NULL,"D3D Tutorial", NULL
	};
    RegisterClassEx(&wc);

    // Create the application's window
    HWND win = CreateWindow("D3D Tutorial", "D3D Tutorial 06: Meshes",
		WS_OVERLAPPEDWINDOW, 100, 100, width, height, GetDesktopWindow(), NULL,
		wc.hInstance, NULL);

	if(!init_d3d(win)) {
		return 1;
	}

	if(!init_geom()) {
		return 1;
	}
	
	ShowWindow(win, SW_SHOWDEFAULT);
	UpdateWindow(win);

	MSG msg; 
	memset(&msg, 0, sizeof msg);

	while(msg.message != WM_QUIT) {
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		render(win);
    }
	
	UnregisterClass("D3D Tutorial", wc.hInstance);
    return 0;
}


bool init_d3d(HWND win)
{
    if(!(d3d = Direct3DCreate9(D3D_SDK_VERSION))) {
        return false;
	}

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp;
	memset(&d3dpp, 0, sizeof d3dpp);
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    // Create the D3DDevice
    if(FAILED(d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, win,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &d3ddev)))
    {
        return false;
    }

	//Turn off culling, so we see the front and back of the triangle
    d3ddev->SetRenderState(D3DRS_LIGHTING, FALSE);
    // Turn on the zbuffer
    d3ddev->SetRenderState(D3DRS_ZENABLE, TRUE);
	d3ddev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    d3ddev->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

	for(int i=0; sdr_files[i]; i++) {
		if(!(vsdr[i] = load_shader(sdr_files[i]))) {
			return false;
		}
		sdr_count++;
	}

	if(!(vsdr_cube = load_shader("cube.vs"))) {
		return false;
	}

	if(!(psdr_refl = load_pixel_shader("refl.ps"))) {
		return false;
	}
	if(!(psdr_refr = load_pixel_shader("refr.ps"))) {
		return false;
	}

	// post effects
	if(!(distort = load_pixel_shader("distort.ps"))) {
		return false;
	}
	if(!(blur = load_pixel_shader("blur.ps"))) {
		return false;
	}

    return true;
}

bool init_geom()
{
	LPD3DXBUFFER pD3DXMtrlBuffer;
	IDirect3DVertexBuffer9 *pMeshSourceVB;
	IDirect3DIndexBuffer9 *pMeshSourceIB;
	D3DVERTEX *pSrc;
	D3DVERTEX *pDst;

	if(FAILED(D3DXCreateTextureFromFile(d3ddev, "chrome.bmp", &tex))) {
		MessageBox(NULL, "Texture Load Problem", "error", MB_OK);
		return false;
	}

	if(FAILED(D3DXCreateTextureFromFile(d3ddev, "ramp.bmp", &tex_ramp))) {
		MessageBox(NULL, "Failed to load ramp texture", "error", MB_OK);
		return false;
	}

	if(FAILED(D3DXCreateTextureFromFile(d3ddev, "refmap1.jpg", &tex_env))) {
		MessageBox(NULL, "Failed to load spherical env map", "error", MB_OK);
		return false;
	}

	if(FAILED(D3DXCreateTextureFromFile(d3ddev, "env.jpg", &tex_back))) {
		MessageBox(NULL, "Failed to load backdrop", "error", MB_OK);
		return false;
	}

	if(FAILED(D3DXCreateCubeTextureFromFile(d3ddev, "cubemap.dds", &tex_cube))) {
		MessageBox(NULL, "Failed to load cubemap", "error", MB_OK);
		return false;
	}

	if(FAILED(D3DXLoadMeshFromX("skull.x", D3DXMESH_SYSTEMMEM, d3ddev, NULL, &pD3DXMtrlBuffer, NULL,
			&material_count, &mesh))) {
		MessageBox(NULL, "Failed to load skull mesh", "error", MB_OK);
		return false;
	}

	if(!create_render_target(width, height, &rtex, &rtex_dsurf)) {
		return false;
	}

	d3ddev->SetFVF(D3DFVF_D3DVERTEX);
	num_verts = mesh->GetNumVertices();
	num_faces = mesh->GetNumFaces();

	//Clone the mesh to set the FVF
	LPD3DXMESH pTempSysMemMesh = NULL;
	
	if(FAILED(mesh->CloneMeshFVF(D3DXMESH_SYSTEMMEM, D3DFVF_D3DVERTEX, d3ddev, &pTempSysMemMesh))) {
		MessageBox(NULL, "Mesh clone problem", "error", MB_OK);
	}
	mesh->Release();

	mesh = pTempSysMemMesh;

	//Compute normals in case the meshes have them
	if(mesh) {
		D3DXComputeNormals(mesh, NULL);
	}

	if(FAILED(d3ddev->CreateVertexBuffer(num_verts * sizeof(D3DVERTEX), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &mesh_vbuf, NULL))) {
		MessageBox(NULL, "Vertex buffer create problem", "error", MB_OK);
		return false;
	}

	if(FAILED(d3ddev->CreateIndexBuffer(num_faces * 3 * sizeof(WORD), D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, &mesh_ibuf, NULL))) {
		MessageBox(NULL, "Index buffer create problem", "error", MB_OK);
		return false;
	}
	mesh->GetVertexBuffer(&pMeshSourceVB);
	mesh_vbuf->Lock( 0, 0, (void**)&pDst, 0 );
	pMeshSourceVB->Lock( 0, 0, (void**)&pSrc, 0 );
	memcpy( pDst, pSrc, num_verts * sizeof(D3DVERTEX) );
	mesh_vbuf->Unlock();
	pMeshSourceVB->Unlock();
	pMeshSourceVB->Release();

	mesh->GetIndexBuffer(&pMeshSourceIB);
	mesh_ibuf->Lock( 0, 0, (void**)&pDst, 0 );
	pMeshSourceIB->Lock( 0, 0, (void**)&pSrc, 0 );
	memcpy( pDst, pSrc, num_faces * 3 * sizeof(WORD));
	mesh_ibuf->Unlock();
	pMeshSourceIB->Unlock();
	pMeshSourceIB->Release();

	pD3DXMtrlBuffer->Release();
    return true;
}

IDirect3DVertexShader9 *load_shader(const char *fname)
{
	IDirect3DVertexShader9 *sdr;
	LPD3DXBUFFER code;
	LPD3DXBUFFER errmsg = 0;
	if(FAILED(D3DXAssembleShaderFromFile(fname, NULL, NULL, 0, &code, &errmsg))) {
		MessageBoxA(0, errmsg ? (char*)errmsg->GetBufferPointer() : "vs error", fname, NULL);
		return 0;
	}
	d3ddev->CreateVertexShader((DWORD*)code->GetBufferPointer(), &sdr);
	code->Release();
	return sdr;
}

IDirect3DPixelShader9 *load_pixel_shader(const char *fname)
{
	IDirect3DPixelShader9 *sdr;
	LPD3DXBUFFER code;
	LPD3DXBUFFER errmsg = 0;
	if(FAILED(D3DXAssembleShaderFromFile(fname, NULL, NULL, 0, &code, &errmsg))) {
		MessageBoxA(0, errmsg ? (char*)errmsg->GetBufferPointer() : "ps error", fname, NULL);
		return 0;
	}
	d3ddev->CreatePixelShader((DWORD*)code->GetBufferPointer(), &sdr);
	code->Release();
	return sdr;
}

void cleanup()
{
	delete [] mesh_materials;

	if(mesh_textures) {
		for(unsigned int i=0; i<material_count; i++) {
			if(mesh_textures[i]) {
				mesh_textures[i]->Release();
			}
		}
		delete [] mesh_textures;
	}
	if(mesh) {
		mesh->Release();
	}

	if(d3ddev) {
		d3ddev->Release();
	}

	if(d3d) {
		d3d->Release();
	}
}

void setup_xform()
{
	// For our world matrix, we will just leave it as the identity
	D3DXMatrixIdentity(&world_mat);
	if(rotate) {
		unsigned int msec = timeGetTime();
		D3DXMatrixRotationYawPitchRoll(&world_mat, msec/1000.0f, msec/700.0f, msec/850.0f);
	}
	d3ddev->SetTransform(D3DTS_WORLD, &world_mat);
  
	D3DXVECTOR3 cpos;
	cpos.x = cos(theta) * sin(phi) * cam_dist;
	cpos.y = cos(phi) * cam_dist;
	cpos.z = sin(theta) * sin(phi) * cam_dist;

	// setup view matrix
	D3DXMatrixLookAtLH(&view_mat, &cpos,
		&D3DXVECTOR3(0.0f, 0.0f, 0.0f), &D3DXVECTOR3(0.0f, 1.0f, 0.0f));
	d3ddev->SetTransform(D3DTS_VIEW, &view_mat);
  
	// set projection matrix
	D3DXMatrixPerspectiveFovLH(&proj_mat, D3DX_PI/4, 1.0f, 1.0f, 1000.0f);
	d3ddev->SetTransform(D3DTS_PROJECTION, &proj_mat);
}


void render(HWND win)
{
	float t = global_time = timeGetTime() / 1000.0f;

	set_render_target(rtex, rtex_dsurf);

	d3ddev->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x0a0a0a, 1.0f, 0);

	d3ddev->BeginScene();

	d3ddev->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	
	if(sdr_idx >= SDR_REFL_SPH) {
		draw_backdrop(tex_back);
	}

	setup_xform();

	if(sdr_idx >= SDR_REFL_CUBE) {
		draw_cube();
	}

	D3DXMATRIX mvp, mv, world, inv_world, view, inv_view;
	D3DXMatrixMultiply(&mv, &world_mat, &view_mat);
	D3DXMatrixMultiply(&mvp, &mv, &proj_mat);
	D3DXMatrixTranspose(&mv, &mv);
	D3DXMatrixTranspose(&mvp, &mvp);

	d3ddev->SetVertexShaderConstantF(0, (float*)&mvp, 4);
	d3ddev->SetVertexShaderConstantF(10, (float*)&mv, 4);

	D3DXMatrixInverse(&inv_view, 0, &view_mat);
	D3DXMatrixTranspose(&inv_view, &inv_view);

	d3ddev->SetVertexShaderConstantF(42, (float*)&inv_view, 4);

	float color[4] = { 0.8f, 0.8f, 0.8f, 1.0f };
	d3ddev->SetVertexShaderConstantF(4, (float*)&color, 1);

	float fLightPosition[3] = {-40.0f, 30.0f, -70.0f};
	d3ddev->SetVertexShaderConstantF(5, (float*)&fLightPosition,1);

	float color1[4] = { 1.0f, 0.0f, 0.0f, 1.0f };
	d3ddev->SetVertexShaderConstantF(6, (float*)&color1, 1);

	POINT ptCursor;
		GetCursorPos( &ptCursor );
		ScreenToClient( win, &ptCursor );
		D3DXVECTOR3 v;
		v.x = -( ( ( 4.0f * ptCursor.x ) / 600 ) - 1 ) / proj_mat._11;
		v.y = ( ( ( 4.0f * ptCursor.y ) / 600 ) - 1 ) / proj_mat._22;
		v.z =  1.0f;


	float fLightPosition1[3] = {v.x, v.y, v.z};
	d3ddev->SetVertexShaderConstantF(7, (float*)fLightPosition1,1);

	float usfconst[4] = {t, sin(t), fabs(sin(t * 2.0)) * 0.25, 1.5 + fabs(sin(timeGetTime()/1500.0f)/4.0f)};
	d3ddev->SetVertexShaderConstantF(9, (float*)usfconst, 1);

	d3ddev->SetFVF(D3DFVF_D3DVERTEX);

	draw_skull();

	set_render_target(0, 0);
	d3ddev->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x0a0a0a, 1.0f, 0);

	draw_backdrop(rtex, post_sdr);
	
	d3ddev->EndScene();
	d3ddev->Present(0, 0, 0, 0);
}

void draw_skull()
{
	d3ddev->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	d3ddev->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	d3ddev->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	d3ddev->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	switch(sdr_idx) {
	case SDR_TOON:
		d3ddev->SetTexture(0, tex_ramp);
		d3ddev->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
		d3ddev->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
		break;

	case SDR_REFL_SPH:
		d3ddev->SetTexture(0, tex_env);
		break;

	case SDR_REFL_CUBE:
		d3ddev->SetPixelShader(psdr_refl);
		d3ddev->SetTexture(0, tex_cube);
		break;

	case SDR_REFR:
		d3ddev->SetPixelShader(psdr_refr);
		d3ddev->SetTexture(0, tex_cube);
		d3ddev->SetTexture(1, tex_cube);
		break;

	default:
		d3ddev->SetTexture(0, 0);
	}
	
	d3ddev->SetVertexShader(vsdr[sdr_idx]);

	d3ddev->SetStreamSource(0, mesh_vbuf, 0, sizeof(D3DVERTEX));
	d3ddev->SetIndices(mesh_ibuf);
	d3ddev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, num_verts, 0, num_faces);

	d3ddev->SetTexture(0, 0);
	d3ddev->SetVertexShader(0);
	d3ddev->SetPixelShader(0);
}

void draw_backdrop(IDirect3DTexture9 *tex, IDirect3DPixelShader9 *ps)
{
	const D3DVERTEX vert[] = {
		{D3DXVECTOR3(-1, 1, 0),	D3DXVECTOR3(0, 0, 0),	0, 0},
		{D3DXVECTOR3(1, 1, 0),	D3DXVECTOR3(0, 0, 0),	1, 0},
		{D3DXVECTOR3(1, -1, 0),	D3DXVECTOR3(0, 0, 0),	1, 1},
		{D3DXVECTOR3(-1, -1, 0),D3DXVECTOR3(0, 0, 0),	0, 1}
	};
	const unsigned short idx[] = {0, 1, 3, 1, 2, 3};

	D3DXMATRIX idmat;
	D3DXMatrixIdentity(&idmat);
	d3ddev->SetTransform(D3DTS_WORLD, &idmat);
	d3ddev->SetTransform(D3DTS_VIEW, &idmat);
	d3ddev->SetTransform(D3DTS_PROJECTION, &idmat);

	d3ddev->SetRenderState(D3DRS_ZWRITEENABLE, 0);

	d3ddev->SetPixelShader(ps);
	if(ps) {
		float tval[] = {global_time, sin(global_time), cos(global_time), 1.0};
		d3ddev->SetPixelShaderConstantF(0, tval, 1);

		float freq[] = {6, 8, 10, 1};
		//float freq[] = {2, 4, 8, 1};
		//float freq[] = {0.5, 1, 2, 1};
		d3ddev->SetPixelShaderConstantF(1, freq, 1);

		float ampl[] = {1, 0.5, 0.25, 1.0 / 16.0};
		d3ddev->SetPixelShaderConstantF(2, ampl, 1);

		float c1[] = {D3DSINCOSCONST1};
		float c2[] = {D3DSINCOSCONST2};
		d3ddev->SetPixelShaderConstantF(5, c1, 1);
		d3ddev->SetPixelShaderConstantF(6, c2, 1);
	}

	d3ddev->SetVertexShader(0);
	d3ddev->SetTexture(0, tex);
	d3ddev->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	d3ddev->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	d3ddev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, sizeof vert / sizeof *vert,
		2, idx, D3DFMT_INDEX16, vert, sizeof *vert);

	d3ddev->SetTexture(0, 0);
	d3ddev->SetPixelShader(0);

	d3ddev->SetRenderState(D3DRS_ZWRITEENABLE, 1);
}

D3DVERTEX cube_verts[] = {
	{D3DXVECTOR3(-1, 1, -1),	D3DXVECTOR3(-1, 1, -1),		0, 0},
	{D3DXVECTOR3(1, 1, -1),		D3DXVECTOR3(1, 1, -1),		0, 0},
	{D3DXVECTOR3(1, -1, -1),	D3DXVECTOR3(1, -1, -1),		0, 0},
	{D3DXVECTOR3(-1, -1, -1),	D3DXVECTOR3(-1, -1, -1),	0, 0},
	{D3DXVECTOR3(-1, 1, 1),		D3DXVECTOR3(-1, 1, 1),		0, 0},
	{D3DXVECTOR3(1, 1, 1),		D3DXVECTOR3(1, 1, 1),		0, 0},
	{D3DXVECTOR3(1, -1, 1),		D3DXVECTOR3(1, -1, 1),		0, 0},
	{D3DXVECTOR3(-1, -1, 1),	D3DXVECTOR3(-1, -1, 1),		0, 0}
};

unsigned short cube_idx[] = {
	0, 1, 3, 1, 2, 3,
	2, 1, 5, 5, 6, 2,
	5, 4, 6, 4, 7, 6,
	4, 0, 7, 0, 3, 7,
	0, 4, 5, 0, 5, 1,
	7, 3, 2, 7, 2, 6
};

void draw_cube()
{
	d3ddev->SetPixelShader(0);
	d3ddev->SetVertexShader(vsdr_cube);
	d3ddev->SetTexture(0, tex_cube);
	d3ddev->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	d3ddev->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	D3DXMATRIX mat(
		100, 0, 0, 0,
		0, 100, 0, 0,
		0, 0, 100, 0,
		0, 0, 0, 1
	);
	d3ddev->SetTransform(D3DTS_WORLD, &mat);

	D3DXMATRIX mvp, mv;
	D3DXMatrixMultiply(&mv, &mat, &view_mat);
	D3DXMatrixMultiply(&mvp, &mv, &proj_mat);
	D3DXMatrixTranspose(&mv, &mv);
	D3DXMatrixTranspose(&mvp, &mvp);
	d3ddev->SetVertexShaderConstantF(0, (float*)&mvp, 4);
	d3ddev->SetVertexShaderConstantF(10, (float*)&mv, 4);
	d3ddev->SetVertexShaderConstantF(14, (float*)&mat, 4);

	d3ddev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

	d3ddev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, sizeof cube_verts / sizeof *cube_verts,
		12, cube_idx, D3DFMT_INDEX16, cube_verts, sizeof *cube_verts);

	d3ddev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	d3ddev->SetTexture(0, 0);
	d3ddev->SetVertexShader(0);
}

bool create_render_target(int xsz, int ysz, IDirect3DTexture9 **rt, IDirect3DSurface9 **dsurf)
{
	if(FAILED(d3ddev->CreateTexture(xsz, ysz, 1, D3DUSAGE_RENDERTARGET, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, rt, 0))) {
		MessageBox(0, "failed to create render target", "error", MB_OK);
		return false;
	}
	if(FAILED(d3ddev->CreateDepthStencilSurface(xsz, ysz, D3DFMT_D24S8, D3DMULTISAMPLE_NONE, 0, TRUE, dsurf, 0))) {
		MessageBox(0, "failed to create depth/stencil surface for the render target", "error", MB_OK);
		return false;
	}
	return true;
}

void set_render_target(IDirect3DTexture9 *rt, IDirect3DSurface9 *dsurf)
{
	static IDirect3DSurface9 *back_surf, *back_ds, *prev_surf;

	IDirect3DSurface9 *surf;
	if(!back_surf) {
		d3ddev->GetRenderTarget(0, &back_surf);
		d3ddev->GetDepthStencilSurface(&back_ds);
		prev_surf = back_surf;
	}

	if(rt) {
		rt->GetSurfaceLevel(0, &surf);

		d3ddev->SetRenderTarget(0, surf);
		d3ddev->SetDepthStencilSurface(dsurf);
	} else {
		d3ddev->SetRenderTarget(0, back_surf);
		d3ddev->SetDepthStencilSurface(back_ds);
	}
}

LRESULT WINAPI msg_handler(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam)
{
	int x, y;
	static int prev_x = -1, prev_y;

	switch(msg) {
	case WM_DESTROY:
		cleanup();
		PostQuitMessage(0);
		break;

	case WM_LBUTTONDOWN:
		prev_x = lparam & 0xffff;
		prev_y = (lparam >> 16) & 0xffff;
		break;

	case WM_LBUTTONUP:
		prev_x = prev_y = -1;
		break;

	case WM_MOUSEMOVE:
		if(prev_x != -1) {
			x = lparam & 0xffff;
			y = (lparam >> 16) & 0xffff;
			theta += (float)(prev_x - x) * 0.01;
			phi += (float)(prev_y - y) * 0.01;

			if(phi < 0.01) phi = 0.01;
			if(phi >= M_PI) phi = M_PI - 0.01;

			prev_x = x;
			prev_y = y;
		}
		break;

	case WM_MOUSEWHEEL:
		if((int)wparam < 0) {
			cam_dist += 1;
		} else {
			cam_dist -= 1;
		}
		break;

	case WM_KEYDOWN:
		switch(wparam) {
		case VK_ESCAPE:
			DestroyWindow(win);
			break;

		case 'd':
		case 'D':
			if(post_sdr == distort) {
				post_sdr = 0;
			} else {
				post_sdr = distort;
			}
			break;

		case 'b':
		case 'B':
			if(post_sdr == blur) {
				post_sdr = 0;
			} else {
				post_sdr = blur;
			}
			break;

		case 'r':
		case 'R':
			rotate = !rotate;
			break;

		case ' ':
			sdr_idx = (sdr_idx + 1) % sdr_count;
			break;

		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			sdr_idx = wparam - '1';
			break;

		case VK_LEFT:
			theta += 0.5;
			break;

		case VK_RIGHT:
			theta -= 0.5;
			break;
		}
		break;

	default:
		return DefWindowProc(win, msg, wparam, lparam);
    }

	return 0;
}